/*****************************************************************************
 * Copyright (C) 2004
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/
using System;
using System.Collections.Specialized;
using iTunesLib;
using System.IO;
using System.Collections;
using System.Threading;

namespace iOpener
{
	/// <summary>
	/// This is the iTunes interface class for iOpener.
	/// </summary>
	public class iOpener 
	{
		private static String				decryptedSuffix = "m4a";
		private static String				encryptedSuffix = "m4p";
		private static String				protectedKindAsString = "Protected AAC audio file";
		private HybridDictionary			modifiedTracks = new HybridDictionary();	// a collection object that will hold decrypted tracks for library updates after library iteration
		private iTunesApp					app;										// the iTunes COM object
		private IITLibraryPlaylist			mainLib;									// the main iTunes library
		private LogHandler					lh;											// the delegate for logging events
		private iOpenerApplicationContext	applicationContext;							// handle to controller app context

		/// <summary>
		/// This declaration allows users of this class to specify a delegate method to be used
		/// for logging events in this class' methods.  The transient parameter means the log message need not be
		/// displayed for more than a moment.
		/// </summary>
		public delegate void LogHandler (string message, bool transient);

		/// <summary>
		/// Constructor requires an iTunesApp COM object to work on
		/// </summary>
		/// <param name="app"></param>
		public iOpener(iTunesApp app, iOpenerApplicationContext applicationContext)
		{
			this.app = app;
			this.applicationContext = applicationContext;
			mainLib = this.app.LibraryPlaylist;
		}

		/// <summary>
		/// This method dispatches to the appropriate private methods to decrypt and update
		/// the main iTunes library.
		/// </summary>
		/// <param name="lh"></param>
		/// <param name="shouldCopyFiles"></param>
		public void DecryptLibrary(Object state)
		{
			DecryptLibraryParameters p = (DecryptLibraryParameters)state;
			this.lh = p.GetLogHandler();
			bool rv = DecryptAllTracks(p.ShouldCopyFiles());
			if (lh != null) lh("Finished", false);
			if (rv) UpdateLibrary(lh);
		}

		/// <summary>
		/// This method copies and decrypts a single track.  It must be passed an encrypted 
		/// track. It will delete the encrypted track from the filesystem unless shouldCopyFiles
		/// is true.  If the user's keys are not available, it will raise a KeysNotFoundException
		/// which should be resumable if I do this right.
		/// </summary>
		/// <param name="track"></param>
		/// <param name="shouldCopyFiles"></param>
		public bool DecryptTrack(IITTrack track, bool shouldCopyFiles, LogHandler lh) 
		{
			String name = track.Name;
			int a, b, c, d;
			track.GetITObjectIDs(out a, out b, out c, out d);
			IITFileOrCDTrack t = (IITFileOrCDTrack) app.GetITObjectByID(a, b, c, d);
			String decryptedFilename = Path.GetDirectoryName(t.Location) + "\\" +
				Path.GetFileNameWithoutExtension(t.Location) + "." +
				decryptedSuffix;	
			try
			{
				File.Copy(t.Location, decryptedFilename, true);
				DeDRMS.Decrypt(new String[1] {decryptedFilename});
				modifiedTracks.Add(t, decryptedFilename);
				if (lh != null) lh("iOpened \"" + track.Artist + ", " + track.Name + "\"", false);
				if (!shouldCopyFiles) 
				{
					File.Delete(t.Location);
					if (lh != null) lh("Deleted original \"" + track.Artist + ", " + track.Name + "\"", false);
				}
			}
			catch (KeyNotFoundException knf)
			{
				try 
				{
					DictionaryEntry de = applicationContext.GetItmsCredentials();
					if (de.Key == null)
					{
						throw;
					}
					int keyType = applicationContext.GetItmsLoginType();
					FairKeys.GetKeys(new String[3] {(String)de.Key, (String)de.Value, Convert.ToString(keyType)});
					File.Delete(decryptedFilename);
					DecryptTrack(track, shouldCopyFiles, lh);
					return true;
				}
				catch (Exception e) 
				{
					if (lh != null)
					{
						lh("iOpener couldn't retrieve your iTunes Music Store key.", false);
						lh("Operation aborted.", false);
					}
					throw new KeyRetrievalException();
				}
			}
			catch( Exception e )
			{
				if (lh != null) lh("Cannot iOpen track: \"" + name + "\"", false);
				Console.WriteLine(e.Message);
				throw;
			}
			return true;
		}

		/// <summary>
		/// This method replaces encrypted tracks in user playlists with the
		/// decrypted tracks from the most recent decryption operation (which would have
		/// populated the modifiedTracks inst var).  It then does the same for the library
		/// playlist, thereby completely removing the encrypted track from iTunes.
		/// </summary>
		public void UpdateLibrary(LogHandler lh)
		{
			HybridDictionary updatedModifiedTracks = new HybridDictionary(modifiedTracks.Count);
			foreach (DictionaryEntry de in modifiedTracks)
			{
				IITOperationStatus status;
				status = app.LibraryPlaylist.AddFile((String)de.Value);
				while (status.InProgress);
				IITTrackCollection addedTracks = status.Tracks;
				// Assumes only one track added at a time.  Probably safe here.
				IITTrack newTrack = addedTracks[1];
				CopyFileAttributes((IITTrack)de.Key, newTrack);
				updatedModifiedTracks.Add(de.Key, newTrack);
			}
			modifiedTracks = updatedModifiedTracks;  // typed languages suck

			IITPlaylist playlist;
			IITSource librarySource = app.LibrarySource;
			for (int i=1; i<=librarySource.Playlists.Count; i++)
			{
				playlist = librarySource.Playlists[i];
				if (playlist.Kind == ITPlaylistKind.ITPlaylistKindUser && !((IITUserPlaylist)playlist).Smart)
				{
					UpdatePlaylist(playlist, lh);
				}
			}
			playlist = app.LibraryPlaylist;
			foreach (DictionaryEntry de in modifiedTracks)
			{
				IITFileOrCDTrack oldTrack = (IITFileOrCDTrack)de.Key;
				oldTrack.Delete();
			}
				
		}

		/// <summary>
		/// This method simply indicates whether a track is encrypted or not
		/// </summary>
		/// <param name="track"></param>
		/// <returns></returns>
		public bool IsProtected(IITTrack track)
		{
			IITFileOrCDTrack t = (IITFileOrCDTrack)track;
			return Path.GetExtension(t.Location) == "." + encryptedSuffix;
		}

		/// <summary>
		/// This method accepts a single playlist and iterates over encrypted tracks,
		/// replacing them with the decrypted one.
		/// </summary>
		/// <param name="playlist"></param>
		/// <param name="lh"></param>
		private void UpdatePlaylist(IITPlaylist playlist, LogHandler lh)
		{
			ArrayList tracksToAdd = new ArrayList();
			bool playlistUpdated = false;
			for (int i=1; i<=playlist.Tracks.Count; i++)
			{
				foreach (DictionaryEntry de in modifiedTracks)
				{
					IITFileOrCDTrack oldTrack = (IITFileOrCDTrack)de.Key;
					IITFileOrCDTrack newTrack = (IITFileOrCDTrack)de.Value;
					if (!(playlist.Tracks[i].TrackDatabaseID == oldTrack.TrackDatabaseID)) continue;
					playlistUpdated = true;
					tracksToAdd.Add(newTrack);
				}
			}
			foreach (Object track in tracksToAdd)
			{
				Object o = (Object)track;
				((IITUserPlaylist)playlist).AddTrack(ref o);
			}
			if (lh != null && playlistUpdated) lh("Updated playlist \"" + playlist.Name + "\"", false);
		}
		
		/// <summary>
		/// This method iterates over the entire main iTunes library and decrypts all encrypted
		/// tracks (deleting the original track from the filesystem if so instructed) and places
		/// the list of decrypted tracks in a dictionary for library updating once iteration
		/// over the main library is complete
		/// </summary>
		/// <param name="shouldCopyFiles"></param>
		private bool DecryptAllTracks(bool shouldCopyFiles) 
		{
			IITTrackCollection tracks = mainLib.Tracks;
			
			bool encryptedTracks = false;

			for (int i=1; i<=tracks.Count; i++) 
			{
				lh("Examining track: \"" + tracks[i].Artist + ", " + tracks[i].Name + "\"", true);
				String kind = tracks[i].KindAsString;
				if (IsProtected(tracks[i])) 
				{
					encryptedTracks = true;
					try 
					{
						bool rv = DecryptTrack(tracks[i], shouldCopyFiles, lh);
						if (!rv) i--;
					}
					catch (KeyRetrievalException kr)
					{
						return false;
					}
					catch (Exception e)
					{
						// Couldn't decrypt that particular track.  Just move on.
						// TODO: re-examine this whole return/exception mechanism
					}
				}
			}
			if (encryptedTracks) 
			{
				return true;
			} 
			else 
			{
				if (lh != null) lh("No iTunes to iOpen", false);
				return false;
			}
		}
		
		/// <summary>
		/// This method copies attributes (name, rating, etc) from an encrypted track to the
		/// decrypted track that will replace it in the library and playlists
		/// </summary>
		/// <param name="oldTrack"></param>
		/// <param name="newTrack"></param>
		private void CopyFileAttributes(IITTrack oldTrack, IITTrack newTrack) 
		{
			//TODO: figure out how to set these
			//newTrack.DateAdded = oldTrack.DateAdded;
			//newTrack.ModificationDate = oldTrack.ModificationDate;
			newTrack.Enabled = oldTrack.Enabled;
			if (newTrack.EQ != null) 
			{
				newTrack.EQ = oldTrack.EQ;
			}
		
			newTrack.PlayedCount = oldTrack.PlayedCount;
			newTrack.PlayedDate = oldTrack.PlayedDate;
			newTrack.Rating = oldTrack.Rating;
			newTrack.VolumeAdjustment = oldTrack.VolumeAdjustment;
		}
	
	}
}