/*****************************************************************************
 * Copyright (C) 2004
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/
using System;
using System.Drawing;
using System.ComponentModel;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using System.Data;
using System.Collections;
using System.Management;
using System.Threading;
using iTunesLib;
using Microsoft.Win32;

// TODO: kill annoying dialog when trying to stop iTunes and COM is still hooked up?
// TODO: show main window and log when auto-decrypting
// TODO: clean up getting keys functionality & exception
// TODO: i18n

namespace iOpener
{
	/// <summary>
	/// iOpenerApplicationContext.
	/// This class has several jobs:
	///		- Create the NotifyIcon UI
	///		- Manage the iOpenerForm that pops up
	///		- Determines when the Application should exit
	/// </summary>
	public class iOpenerApplicationContext : ApplicationContext
	{
		private static	iOpenerApplicationContext	applicationContext;				// the application context that controls this form
		private static iTunesApp					app;							// the iTunes COM object
		private static _IiTunesEvents_OnDatabaseChangedEventEventHandler
													databaseChangedEventHandler;	// handler for iTunes lib database adds/deletes/etc.
		private static _IiTunesEvents_OnQuittingEventEventHandler
													quittingEventHandler;			// handler for user-selected quit of iTunes

		private IContainer							components;						// a list of components to dispose when the context is disposed
		private NotifyIcon							iOpenerNotifyIcon;				// the icon that sits in the system tray
		private ContextMenu							iOpenerNotifyIconContextMenu;	// the context menu for the notify icon
		private MenuItem							exitContextMenuItem;			// exit menu command for context menu 
		private MenuItem							showContextMenuItem;			// open menu command for context menu 	
		private MenuItem							showPrefsContextMenuItem;		// show prefs menu command for context menu 
		private iOpenerForm							mainForm;						// the main iOpener form
		private ManagementEventWatcher				iTunesStartupWatcher;			// WMI watcher for the iTunes.exe process
		private bool								shouldCopyFiles = true;			// global pref for whether DRM-ed track should be copied before decrypting
		private bool								shouldAutoDecrypt = true;		// global pref for whether to sit in tray and decrypt new files
		private String								registryKeyName = "Software\\iOpener";	// location in the registry for storing prefs
		private DictionaryEntry						itmsCredentials;				// username and password for retrieval of iTMS key
		private int									itmsLoginType;					// account type (Apple, AOL currently) for the iTMS


		/// <summary>
		/// This class should be created and passed into Application.Run( ... )
		/// </summary>
		public iOpenerApplicationContext()
		{
			// create the notify icon and it's associated context menu
			InitializeContext();
			InitializeItunesApp();
			InitializeProcessNotifier();
			LoadRegistryPrefs();
		}

		/// <summary>
		/// This method returns the iTunesApp COM object, initializing it if necessary
		/// </summary>
		/// <returns></returns>
		public IiTunes GetItunesApp() 
		{
			InitializeItunesApp();
			return app;
		}

		/// <summary>
		/// This method lazy initializes the iTunesApp COM object and registers COM
		/// event handlers
		/// </summary>
		public void InitializeItunesApp()
		{
			if (app == null) 
			{
				app = new iTunesAppClass();
				databaseChangedEventHandler = new _IiTunesEvents_OnDatabaseChangedEventEventHandler(DatabaseChangedEventHandler);
				app.OnDatabaseChangedEvent += databaseChangedEventHandler;
				quittingEventHandler = new _IiTunesEvents_OnQuittingEventEventHandler(QuittingEventHandler);
				app.OnQuittingEvent += quittingEventHandler;
			}
		}

		/// <summary>
		/// This COM event handler detects changes to the iTunes database (adds/deletes/etc)
		/// and passes changed encrypted tunes off to the iOpener class for decryption and
		/// library updates
		/// </summary>
		/// <param name="deletedObjects"></param>
		/// <param name="changedObjects"></param>
		public void DatabaseChangedEventHandler(Object deletedObjects, Object changedObjects) 
		{
			if (!this.shouldAutoDecrypt) return;
			this.InitializeItunesApp();
			Array co = (Array)changedObjects;
			ArrayList seenObjects = new ArrayList();
			iOpener io = new iOpener(app, this);
			iOpener.LogHandler lh = null;
			if (mainForm != null) lh = mainForm.GetLogHandler();
			for (int i=0; i<co.GetLength(0); i++) 
			{
				int sourceID = (int)co.GetValue(i, 0);
				int playlistID = (int)co.GetValue(i, 1);
				int trackID = (int)co.GetValue(i, 2);
				int databaseID = (int)co.GetValue(i, 3);

				IITObject changedObject =
					app.GetITObjectByID(sourceID, playlistID, trackID, databaseID);

				if (trackID != 0 && (io.IsProtected((IITTrack)changedObject)) && !seenObjects.Contains(databaseID))
				{
					io.DecryptTrack((IITTrack)changedObject, this.shouldCopyFiles, lh);
					seenObjects.Add(databaseID);
				}
			}
			if (seenObjects.Count != 0) io.UpdateLibrary(lh);
		}

		/// <summary>
		/// This COM event handler releases the iTunes COM object for garbage collection since
		/// the application is going bye-bye anyway.
		/// </summary>
		public void QuittingEventHandler()
		{
			app = null;
		}

		/// <summary>
		/// This method displays and gets values from the prefs dialog
		/// </summary>
		public void ShowPrefs()
		{
			PrefsDialog prefsDialog = new PrefsDialog(this.shouldCopyFiles, this.shouldAutoDecrypt);
			if (prefsDialog.ShowDialog() == DialogResult.OK)
			{
				this.shouldCopyFiles = prefsDialog.ShouldCopyFiles();
				this.shouldAutoDecrypt = prefsDialog.ShouldAutoDecrypt();
				SetRegistryPrefs();
			}
			prefsDialog.Dispose();
		}

		/// <summary>
		/// Getter method for the shouldCopyFiles boolean pref
		/// </summary>
		public bool ShouldCopyFiles()
		{
			return this.shouldCopyFiles;
		}

		/// <summary>
		/// Getter method for the shouldAutoDecrypt boolean pref
		/// </summary>
		public bool ShouldAutoDecrypt()
		{
			return this.shouldAutoDecrypt;
		}
		
		/// <summary>
		/// This is the public interface method to get credentials which dispatches
		/// to the GetCredentials method if necessary
		/// </summary>
		/// <returns></returns>
		public DictionaryEntry GetItmsCredentials()
		{
			if (this.itmsCredentials.Key == null) GetCredentials();
			return this.itmsCredentials;
		}

		/// <summary>
		/// This is the public interface method to get the type of the iTMS login (Apple, AOL currently)
		/// </summary>
		/// <returns></returns>
		public int GetItmsLoginType()
		{
			return this.itmsLoginType;
		}

		/// <summary>
		/// When the application context is disposed, dispose things like the notify icon.
		/// </summary>
		/// <param name="disposing"></param>
		protected override void Dispose( bool disposing )
		{
			if( disposing )
			{
				if (components != null) 
				{
					components.Dispose();
				}
				app = null;
				iTunesStartupWatcher.Stop();
			}		
		}		

		/// <summary>
		/// If we are presently showing a mainForm, clean it up.
		/// </summary>
		protected override void ExitThreadCore()
		{
			if (mainForm != null) 
			{
				mainForm.Close();
			}
			base.ExitThreadCore ();
		}

		/// <summary>
		/// Create the NotifyIcon UI, the ContextMenu for the NotifyIcon and an Exit menu item. 
		/// </summary>
		private void InitializeContext() 
		{
			this.components = new System.ComponentModel.Container();
			this.iOpenerNotifyIcon = new System.Windows.Forms.NotifyIcon(this.components);
			this.iOpenerNotifyIconContextMenu = new System.Windows.Forms.ContextMenu();
			this.showContextMenuItem = new System.Windows.Forms.MenuItem();
			this.showPrefsContextMenuItem = new System.Windows.Forms.MenuItem();
			this.exitContextMenuItem = new System.Windows.Forms.MenuItem();
		
			// 
			// iOpenerNotifyIcon
			// 
			this.iOpenerNotifyIcon.ContextMenu = this.iOpenerNotifyIconContextMenu;
			this.iOpenerNotifyIcon.DoubleClick += new System.EventHandler(this.iOpenerNotifyIcon_DoubleClick);
			this.iOpenerNotifyIcon.Icon = new Icon(typeof(iOpenerApplicationContext), "iOpener.ico");
			this.iOpenerNotifyIcon.Text = "iOpener";
			this.iOpenerNotifyIcon.Visible = true;

			// 
			// iOpenerNotifyIconContextMenu
			// 
			this.iOpenerNotifyIconContextMenu.MenuItems.AddRange(new MenuItem[] { showContextMenuItem, showPrefsContextMenuItem, exitContextMenuItem });
			
			// 
			// showContextMenuItem
			// 
			this.showContextMenuItem.Index = 0;
			this.showContextMenuItem.Text = "&Show iOpener";
			this.showContextMenuItem.DefaultItem = true;
			this.showContextMenuItem.Click += new System.EventHandler(this.showContextMenuItem_Click);

			//
			// showPrefsContextMenuItem
			//
			this.showPrefsContextMenuItem.Index = 1;
			this.showPrefsContextMenuItem.Text = "&Prefs";
			this.showPrefsContextMenuItem.Click += new System.EventHandler(this.showPrefsMenuItem_Click);
			// 
			// exitContextMenuItem
			// 
			this.exitContextMenuItem.Index = 2;
			this.exitContextMenuItem.Text = "&Exit";
			this.exitContextMenuItem.Click += new System.EventHandler(this.exitContextMenuItem_Click);
		}


		/// <summary>
		/// This method initalizes a WMI handler to watch for iTunes being launched
		/// </summary>
		private void InitializeProcessNotifier()
		{
			WqlEventQuery iTunesStartupEventQuery =
				new WqlEventQuery("SELECT * FROM __InstanceCreationEvent WITHIN 1 WHERE TargetInstance ISA 'Win32_Process'");
			iTunesStartupWatcher = new ManagementEventWatcher(new ManagementScope(@"root\cimv2"), iTunesStartupEventQuery);
			WMIHandler handler = new WMIHandler(this);
			iTunesStartupWatcher.EventArrived += new EventArrivedEventHandler(handler.Started);
			iTunesStartupWatcher.Start();
		}

		/// <summary>
		/// When the open menu item is clicked, make a call to Show the form.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void showContextMenuItem_Click(object sender, EventArgs e) 
		{
			ShowForm();
		}

		/// <summary>
		/// This method calls the method to display prefs
		/// </summary>
		private void showPrefsMenuItem_Click(object sender, EventArgs e)
		{
			ShowPrefs();
		}

		/// <summary>
		/// This method collects the user's iTMS credentials via a dialog.
		/// </summary>
		private void GetCredentials()
		{
			CredentialsDialog credentialsDialog = new CredentialsDialog();
			if (credentialsDialog.ShowDialog() == DialogResult.OK)
			{
				this.itmsCredentials = credentialsDialog.GetCredentials();
				this.itmsLoginType = credentialsDialog.GetLoginType();
			}
			credentialsDialog.Dispose();
		}

		/// <summary>
		/// This method saves prefs to the registry
		/// </summary>
		private void SetRegistryPrefs()
		{
			RegistryKey rk = Registry.CurrentUser.CreateSubKey(this.registryKeyName);
			rk.SetValue("shouldCopyFiles", this.shouldCopyFiles);
			rk.SetValue("shouldAutoDecrypt", this.shouldAutoDecrypt);
		}

		/// <summary>
		/// This method loads prefs from the registry
		/// </summary>
		private void LoadRegistryPrefs()
		{
			RegistryKey rk = Registry.CurrentUser.CreateSubKey(this.registryKeyName);
			this.shouldCopyFiles = ((String)rk.GetValue("shouldCopyFiles", "True")) == "True";
			this.shouldAutoDecrypt = ((String)rk.GetValue("shouldAutoDecrypt", "True")) == "True";
		}

		/// <summary>
		/// When the exit menu item is clicked, make a call to terminate the ApplicationContext.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void exitContextMenuItem_Click(object sender, EventArgs e) 
		{
			ExitThread();
		}
		
		/// <summary>
		/// When the notify icon is double clicked in the system tray, bring up the iOpener form.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void iOpenerNotifyIcon_DoubleClick(object sender,System.EventArgs e)
		{
			ShowForm();
		}

		/// <summary>
		/// Null out the main form so we know to create a new one.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void mainForm_Closed (object sender, EventArgs e) 
		{
			if (this.shouldAutoDecrypt) 
			{
				this.mainForm.Dispose();
				this.mainForm = null;
			} 
			else 
			{
				ExitThread();
			}
		}

		/// <summary>
		/// This method will either create a new iOpener form or activate the
		/// existing one, bringing the window to front.
		/// </summary>
		private void ShowForm() 
		{
			if (mainForm == null) 
			{
				mainForm = new iOpenerForm(applicationContext);
				mainForm.Show();
                        
				// Hook onto the closed event so we can null out the main form.
				// This avoids reshowing a disposed form.
				mainForm.Closed +=new EventHandler(mainForm_Closed);        
			}
			else 
			{
				// The form is currently visible, go ahead and bring it to the front
				mainForm.Activate();
			}
		}

		static void Main() 
		{
			applicationContext = new iOpenerApplicationContext();
			applicationContext.ShowForm();
			Application.Run(applicationContext);
		}

	}
}
